;***********************************************************************
;***********************************************************************
;***********************************************************************
; B O O T S W I T C H     -     A V R   F I R M W A R E
;***********************************************************************
;***********************************************************************
;***********************************************************************
;
; Written and conceived in 2012/2013/2014 by Julien Thomas (Germany)
; Released under Creative Commons (CC BY-SA) --- see README.TXT
;
;***********************************************************************
;
;
;
; ----------------------------------------------------------------------
; HARDWARE PLATFORM AND PORTS I/O
; ----------------------------------------------------------------------
;
; BootSwitch 4-1 Firmware is for ATtiny2313 (4 channels)
;
; PA0/PA1       -- external chrystal (spare, not in use)
; PA2 (/RESET)  <- /RESET also triggered by Button event (falling edge)
; PD0 (RXD)     -- n.c.
; PD1 (TXD)     -- n.c.
; PD2           <- Relay voltage Ready (High = capacitor charged)
; PD3           -- n.c.
; PD4           <- Button-Input (e.g. original POWER-Button)
; PD5           -> Button-Output (Low-active)
; PD6           -> PC-Power ready (High = regular PC voltage present)
;
; PB0-PB7       -> switching outputs for up to 8 hdd channels
;
; See circuit documentation for further details.
;
; ----------------------------------------------------------------------
; FUSEBITS ATTINY2313
; ----------------------------------------------------------------------
;
; Ext   : $FF
; High  : $C9
; Low   : $62
;
; Use internal RC-oscillator of 4 MHz divided thru 8 ( = 0.5 MHz)
; Use Brown-Out-Detector at 4,3 volts
; Do not allow SPM (no risk of flash corruption)
; Watch-dog Timer always ON
;
; ----------------------------------------------------------------------
; DIRECTIVES, CONSTANTS AND REGISTERS
; ----------------------------------------------------------------------

.nolist

.include "tn2313def.inc"

; SRAM buffers

.equ    LOCK    =       SRAM_START+0    ; lock relay decision
.equ    WSTC    =       SRAM_START+1    ; warmstart/Button counter

; Working registers

.def    tmp     =       r16             ; universal temp registers
.def    rly     =       r17             ; relay to energize
.def    tlo     =       r18             ; delay loop (1/100 seconds)

; Control bits in PORTD

.equ    RLY_RDY =       2               ; Relay-voltage Ready
.equ    BTN_IN  =       4               ; Button-Input
.equ    BTN_OUT =       5               ; Button-Output
.equ    PCP_RDY =       6               ; PC-Power Ready

; ----------------------------------------------------------------------
; FIRMWARE START
; ----------------------------------------------------------------------
;
.cseg
.org $0000
                rjmp    RESET             ; only irq is RESET
                reti
                reti
                reti
                reti
                reti
                reti
                reti
                reti
                reti
                reti
                reti
                reti
                reti
                reti
                reti
                reti
                reti
                reti
                reti
                reti

; ----------------------------------------------------------------------
; RESET
; ----------------------------------------------------------------------
; Initialize Stack, Watchdog and control port bits
;
RESET:
                cli
                ldi     tmp, (RAMEND)   ; ramend low stackpointer
                out     SPL, tmp

                clr     tmp
                out     MCUCR, tmp

                wdr

                in      tmp, MCUSR
                andi    tmp, 0b11110111 ; clear Watchdog Reset-Flag
                out     MCUSR, tmp

                in      tmp, WDTCSR     ; timed sequence watchdog reset
                ori     tmp, 0b00011000 ; set WDCE + WDE
                out     WDTCSR, tmp
                ldi     tmp, 0b00101001 ; set WDE + WDP3 + WDP0
                out     WDTCSR, tmp     ; watchdog timer 8 seconds

                clr     tmp
                out     DDRD, tmp
                out     PORTD, tmp

                cbi     DDRD,  BTN_OUT  ; Button-Output
                sbi     PORTD, BTN_OUT  ; not activated yet

                cbi     DDRD,  BTN_IN   ; Button-Input
                sbi     PORTD, BTN_IN   ; with pull-up

                cbi     DDRD,  PCP_RDY  ; Sense PC power
                cbi     PORTD, PCP_RDY  ; no pull-up

                cbi     DDRD,  RLY_RDY  ; Sense Relay power
                cbi     PORTD, RLY_RDY  ; no pull-up


; ----------------------------------------------------------------------
; PC POWER-ON STATE ( REGULAR OPERATION )
; ----------------------------------------------------------------------
; While PC power is on (regular PC voltage present), Button-pushes
; are rather directly forwarded to the Mainboard to enable normal
; Button-functionality (i.e. PC shut-down, force power-off).
; The portline thats just been activated by the POWERON routine will be
; left in high state, until another Reset, Power-down or Button-event
; occurs. This is only to provide some extra safety in addition to
; the hardware-feedback that normally keeps selected relay energized.

PCPowerOnState:

                sbic    PIND, PCP_RDY   ; if PC-Power is present
                rjmp    PCPOS           ; loop and await Button pushes

PCStbS:
                wdr                     ; keep watchdog alive
                sbis    PIND, BTN_IN    ; but wait until Button released
                rjmp    PCStbS          ; before entering Standby-mode
                rjmp    PCStandbyState
PCPOS:
                wdr

                clr     tmp             ; re-set lock status
                sts     LOCK, tmp       ; to unlocked

                ldi     rly, 254        ; re-set warmstart counter
                sts     WSTC, rly       ; to failsafe initial

                ldi     tlo, (10)       ; insert some debouncing delay
                rcall   Waiths

                sbi     DDRD, BTN_OUT   ; set Button-Output to output

                cbi     DDRD,  BTN_IN   ; Button-Input
                sbi     PORTD, BTN_IN   ; with pull-up

                sbis    PIND, BTN_IN    ; direct
                cbi     PORTD, BTN_OUT  ; forwarding
                sbic    PIND, BTN_IN    ; of Button-Input
                sbi     PORTD, BTN_OUT  ; to Button-Output

                rjmp    PCPowerOnState  ; loop until next shutdown


; ----------------------------------------------------------------------
; PC STANDBY STATE ( WAIT AND COUNT BUTTON EVENTS )
; ----------------------------------------------------------------------
;
PCStandbyState:
                wdr

                ldi     tmp, 0b00000000 ; all PORTB control lines "0"
                out     PORTB, tmp
                ldi     tmp, 0b11111111 ; all PORTB lines active output
                out     DDRB, tmp

                sbi     PORTD, BTN_OUT  ; passify Button-Output
                cbi     DDRD,  BTN_OUT

; ----------------------------------------------------------------------
; ENSURE DEFINED STARTING CONDITIONS OF THE CONTROLLER
; ----------------------------------------------------------------------
;
StartCheck:
                in      tmp, MCUSR      ; check MCU Status Register for
                andi    tmp, 0b00000101 ; Brown-Out or Power-On-Reset
                brne    StartInit       ; if any of those, re-init all

                sbis    PIND, BTN_IN    ; is Button still being pressed
                rjmp    CountButtonPushes  ; this IS a count-up-event!
                                        ; else
                lds     tmp, LOCK       ; check if choice was locked,
                tst     tmp             ; then continue waiting for
                breq    WsButtonCheck   ; changes of Button / PC-power
                                        ; yes, choice was locked
                cbi     PORTD, BTN_IN   ; reconfigure Button-In
                sbi     DDRD,  BTN_IN   ; to static output LOW
                rjmp    RelaySwitch     ; and engage relay switching


; ----------------------------------------------------------------------
; CONTROLLER COLDSTART - INIT SRAM VALUES
; ----------------------------------------------------------------------
;
StartInit:
                wdr
                clr     tmp             ; clear all MCUSR flags
                out     MCUSR, tmp
                clr     tmp             ; clear choice-done-lock
                sts     LOCK, tmp       ; (unlocked)
                ldi     rly, 254        ; set warmstart counter to
                sts     WSTC, rly       ; initial state (uncritical)
                rjmp    RESET

; ----------------------------------------------------------------------
; WARMSTART BUTTON CHECKS - COUNT ON / ENGAGE HDD/POWER SWITCHING
; ----------------------------------------------------------------------
;
WsButtonCheck:

                lds     rly, WSTC
                cpi     rly, 254
                brne    WsButtonReleased     ; ignore 1st Button event

                ldi     rly, 255
                sts     WSTC, rly

WsChangeDetect:
                wdr

                ldi     tlo, (10)       ; debouncing delay
                rcall   Waiths

                sbic    PIND, PCP_RDY   ; probe for PC-Power emerging
                rjmp    PCPowerOnState  ; (you never know...!)

                sbic    PIND, BTN_IN    ; probe for Button pressed
                rjmp    WsChangeDetect  ; and go RESET anyway
                rjmp    RESET


WsButtonReleased:
                wdr

                ldi     tlo, (10)       ; debouncing delay
                rcall   Waiths

                sbic    PIND, PCP_RDY   ; probe for PC-Power emerging
                rjmp    PCPowerOnState

                sbis    PIND, BTN_IN    ; loop until Button released
                rjmp    WsButtonReleased

CountButtonPushes:

                lds     rly, WSTC       ; load counter value from SRAM
                inc     rly             ; increase by 1
                sts     WSTC, rly       ; and immediately save to SRAM

; ----------------------------------------------------------------------
; WAIT FOR NEXT Button PUSH OR TIME-OUT AND PREPARE SWITCHING SEQUENCE
; ----------------------------------------------------------------------

WaitButtonTimeout:

                sbi     PORTD, BTN_IN

                ldi     tmp, 150        ; timeout factor * 10 ms

wtbd1:          ldi     YL, low (1250)  ; timing divider for 10 ms
                ldi     YH, high(1250)  ; fClk / 1000 / 4
                wdr
wtbd2:
                sbiw    YL, 1           ; inner Timeout loop is 10 ms
                brne    wtbd2
                dec     tmp             ; if Timeout not elapsed,
                brne    wtbd3           ; loop on

                ser     tmp             ; Timeout elapsed -
                sts     LOCK, tmp       ; choice being locked now !
                rjmp    RESET

wtbd3:          sbic    PIND, BTN_IN    ; leave if Button been pushed
                rjmp    wtbd1           ; else loop on
                rjmp    RESET           ; just to be sure

; ----------------------------------------------------------------------
; ----------------------------------------------------------------------
; ENGAGE RELAY AND POWER-ON-CYCLE UNDER VALID CONDITIONS
; ----------------------------------------------------------------------
; ----------------------------------------------------------------------
;
RelaySwitch:
                lds     tmp, LOCK       ; LOCK not been set?
                tst     tmp             ;
                brne    RelPCPCheck     ;
                rjmp    RESET           ; Not locked, don't switch!

RelPCPCheck:
                sbic    PIND, PCP_RDY   ; Is PC Power still OFF/Standby?
                rjmp    RESET           ; If NOT, go RESET!

RelCharge:      sbis    PIND, RLY_RDY   ; Relay charge not present?
                rjmp    StartInit       ; Forget it!

                wdr                     ; Okay, we're ready to switch
                lds     rly, WSTC       ; load relay number from SRAM
RelSw0:
                cpi     rly, 0          ; check if Relay 0 selected
                brne    RelSw1
                sbi     PORTB, 0        ; set PORTB.0
                rjmp    POWERON         ; initiate Power-On

RelSw1:
                cpi     rly, 1          ; check if Relay 1 selected
                brne    RelSw2
                sbi     PORTB, 1        ; set PORTB.1 ... and so on
                rjmp    POWERON

RelSw2:
                cpi     rly, 2
                brne    RelSw3
                sbi     PORTB, 2
                rjmp    POWERON

RelSw3:
                cpi     rly, 3
                brne    RelSw4
                sbi     PORTB, 3
                rjmp    POWERON

RelSw4:
                cpi     rly, 4
                brne    RelSw5
                sbi     PORTB, 4
                rjmp    POWERON

RelSw5:
                cpi     rly, 5
                brne    RelSw6
                sbi     PORTB, 5
                rjmp    POWERON

RelSw6:
                cpi     rly, 6
                brne    RelSw7
                sbi     PORTB, 6
                rjmp    POWERON

RelSw7:
                cpi     rly, 7
                brne    RelSwX
                sbi     PORTB, 7
                rjmp    POWERON

RelSwX:
                cpi     rly, 8          ; More than 9 pushes detected?
                brlo    POWERON         ; Do NOT trigger Power-on.
                rjmp    StartInit

; ----------------------------------------------------------------------
; TIMED SEQUENCE OF RELAY ACTIVATION AND POWER-ON
; ----------------------------------------------------------------------
;
POWERON:
                ldi     tlo, (10)       ; allow at least 100 ms for
                rcall   Waiths          ; relay to energize (generous!)

                sbi     DDRD,  BTN_OUT  ; push Button-Out (low-active)
                cbi     PORTD, BTN_OUT  ; (Mainboard Power-On cycle)

                ldi     tlo, (30)       ; wait at least 300 ms
                rcall   Waiths          ; for power supply

                ldi     rly, 254        ; re-set warmstart counter to
                sts     WSTC, rly       ; initial value

; Wait even some more seconds for PC-voltage to appear, or time-out

                ldi     tmp, 100        ; timeout factor 100x
PWRNWait:
                wdr

                ldi     tlo, (5)        ; internal delay 5 hs (50 ms)
                rcall   Waiths

                sbic    PIND, PCP_RDY   ; if PC-voltage detected,
                rjmp    PWRNfinished    ; everything will be fine

                dec     tmp             ; else count down timeout
                brne    PWRNWait        ; sorry, if timeout elapsed ...

; then POWER-ON HAS FAILED!
; Relay will be switched off and we return to Standby-Mode.

                sbi     PORTD, BTN_OUT  ; release Button-Out
                cbi     DDRD,  BTN_OUT  ; and passify its output line

                cbi     DDRD,  BTN_IN   ; re-enable Button-In
                sbi     PORTD, BTN_IN   ; (Button-triggered Resets)

                clr     tmp             ; clear choice-done-lock
                sts     LOCK, tmp       ; (unlocked)

                ldi     tlo, (100)
                rcall   Waiths

                rjmp    PCStandbyState


; POWER-ON SUCCESSFUL! Restore values and enter Power-On-Mode.
; Relay will hold by hardware-feedback AND portline left active HIGH.

PWRNfinished:
                sbi     PORTD, BTN_OUT  ; release Button-Out
                cbi     DDRD,  BTN_OUT  ; and passify its output line

                cbi     DDRD,  BTN_IN   ; re-enable Button-In
                sbi     PORTD, BTN_IN   ; (Button-triggered Resets)

                clr     tmp             ; clear choice-done-lock
                sts     LOCK, tmp       ; (unlocked)

                ldi     tlo, (200)
                rcall   Waiths

                rjmp    PCPowerOnState


; ----------------------------------------------------------------------
; WAIT LOOP 1/100 SECONDS (tlo)
; ----------------------------------------------------------------------
;
Waiths:
                wdr                     ; quiet, doggy!
whs1:
                ldi     YL, low (1250)  ; timer factor
                ldi     YH, high(1250)  ; 500.000 / 1250 / 4
whs2:           sbiw    YL, 1           ; sbiw + brne = 4 clocks
                brne    whs2
                dec     tlo
                brne    whs1
                ret

; ----------------------------------------------------------------------
;
.exit
;
;***********************************************************************
; END OF BOOTSWITCH PROGRAM
;***********************************************************************
